function D = D_ADI(P, N, lb, ub)
% D = D_ADI(P, N, lb, ub)
%
% Computes the ADI measure.
%
% Authors: Johannes W. Kruisselbrink, Michael T.M. Emmerich, Andre H. Deutz

	% If P is empty, then D = inf
	if (isempty(P))
		D = norm(ub - lb);
		return;
	end

	if (N == 2)
		D = D_ADI_voronoi(P, N, lb, ub);
	else
		D = D_ADI_approximate(P, N, lb, ub);
	end

end

function [D] = D_ADI_voronoi(P, N, lb, ub)

	M = size(P,2);

	P_left = [lb(1) - (P(1,:) - lb(1)) ; P(2,:)];
	P_right = [ub(1) + (ub(1) - P(1,:)) ; P(2,:)];
	P_down = [P(1,:) ; lb(2) - (P(2,:) - lb(2))];
	P_up = [P(1,:) ; ub(2) + (ub(2) - P(2,:))];
	P_plus = [P , P_left , P_right , P_down , P_up ];

%	plot(P_plus(1,:), P_plus(2,:), 'kx')

	[VX, VY] = voronoi(P_plus(1,:), P_plus(2,:));
%	h = voronoi(P_plus(1,:), P_plus(2,:));
%	hold on

	total_surface = prod(ub - lb);

	total_volume = 0;
	for i = 1:size(VX,2)
		p1 = [VX(1,i); VY(1,i)];
		p2 = [VX(2,i); VY(2,i)];
		epsi=1e-5;
		if (p1(1) >= lb(1)-epsi && p1(1) <= ub(1)+epsi && p1(2) >= lb(2)-epsi &&  p1(2) <= ub(2)+epsi && ...
		    p2(1) >= lb(1)-epsi && p2(1) <= ub(1)+epsi && p2(2) >= lb(2)-epsi &&  p2(2) <= ub(2)+epsi)

			c = p1 + 0.5 * (p2 - p1);

			distances = sqrt(sum((repmat(c,1,M) - P).^2, 1));
			[sorted_distances, sorted_indexes] = sort(distances,'ascend');

			if (length(distances) > 1 && abs(sorted_distances(1) - sorted_distances(2)) <= epsi) 

				% Nearest point 1
				nep1 = P(:,sorted_indexes(1));

				% Nearest point 2
				nep2 = P(:,sorted_indexes(2));

				% Triangle one
				surface_area_1 = .5 * abs(det([p1', 1 ; p2', 1 ; nep1', 1]));
				avg_height_1 = mean_distance_volume(p1, p2, nep1);
				total_volume = total_volume + surface_area_1 * avg_height_1;
%				fill([nep1(1), p1(1), p2(1)], [nep1(2), p1(2), p2(2)], 'g');

				% Triangle two
				surface_area_2 = .5 * abs(det([p1', 1 ; p2', 1 ; nep2', 1]));
				avg_height_2 = mean_distance_volume(p1, p2, nep2);
				total_volume = total_volume + surface_area_2 * avg_height_2;
%				fill([nep2(1), p1(1), p2(1)], [nep2(2), p1(2), p2(2)], 'g');

			else

				% Nearest point 1
				nep1 = P(:,sorted_indexes(1));

				% Compute surface area triangle one
				surface_area_1 = .5 * abs(det([p1', 1 ; p2', 1 ; nep1', 1]));
				avg_height_1 = mean_distance_volume(p1, p2, nep1);
				total_volume = total_volume + surface_area_1 * avg_height_1;
%				fill([nep1(1), p1(1), p2(1)], [nep1(2), p1(2), p2(2)], 'm');

			end
		end
	end
%	hold off
%	drawnow()

	D = total_volume / prod(ub - lb);

end


function V = mean_distance_volume(a, b, c)
	u = (norm(a-c) + norm(b-c)) / norm(a - b);
	v = (norm(a-c) - norm(b-c)) / norm(a - b);
	if ((u - 1)/(u + 1) > 0)
		V = (norm(a - b) / 6) * (u * (1 + v^2) + (1/2) * (1 - u^2) * (1 - v^2) * log((u - 1)/(u + 1)));
	else
		V = 0;
	end
end


function D = D_ADI_approximate(P, N, lb, ub)

	% Approximate the continuous space by an N-dimensional
	% grid with the following resolution
	resolution = 100;
	m = resolution^N;

	% Get an ND-grid of the specified resolution 
	R = grid_sampling(m, N, lb, ub);

	% Compute the distance of every sample point
	% to P and add it to the volume v
	v = 0;
	for i = 1:m
		v = v + min(sqrt(sum((repmat(R(i,:),size(P,2),1)' - P).^2)));
	end

	% Compute the ADI approximation
	D = v / m;

end
